"""Shared utilities for tests."""

from subprocess import check_call
from tempfile import mkdtemp
import os
import sys

import cbor2

from pycddl import Schema

BSTR_SCHEMA = Schema("object = bstr")

BSTR_1K = cbor2.dumps(b"A" * 1_000)
BSTR_1M = cbor2.dumps(b"A" * 1_000_000)


# Generate the largest bstr in a subprocess, so that we don't increase maxrss
# in this process unnecessarily:
def _make_bstr_100m():
    path = os.path.join(mkdtemp(), "out.cbor")
    check_call(
        [
            sys.executable,
            "-c",
            """\
import cbor2
with open({}, 'wb') as f:
    cbor2.dump(b"A" * 100_000_000, f)
""".format(
                repr(path)
            ),
        ]
    )
    with open(path, "rb") as f:
        return f.read()


BSTR_100M = _make_bstr_100m()
assert len(BSTR_100M) == 100_000_005



"""
Tests for CBOR decoding functionality.
"""

import math

from pycddl import Schema
from hypothesis import given, strategies as st, settings
import cbor2

ANY_SCHEMA = Schema("main = any")

PRIMITIVES = (
    st.none()
    | st.booleans()
    | st.floats(allow_nan=False)
    | st.text()
    | st.integers(max_value=2**63, min_value=-(2**63))
    | st.binary()
)


@given(
    structure=st.recursive(
        PRIMITIVES,
        lambda values: st.lists(values) | st.dictionaries(PRIMITIVES, values),
        max_leaves=10,
    )
)
@settings(max_examples=2_000)
def test_roundtrip_basic_decoding(structure):
    """
    Basic serialized structures can be deserialized by pycddl.
    """
    serialized = cbor2.dumps(structure)
    deserialized = ANY_SCHEMA.validate_cbor(serialized, True)
    assert structure == deserialized


def test_decoding_is_optional():
    """
    If ``validate_cbor`` isn't told to, it doesn't deserialize the CBOR.
    """
    serialized = cbor2.dumps({1: 2})
    assert ANY_SCHEMA.validate_cbor(serialized, False) is None
    assert ANY_SCHEMA.validate_cbor(serialized) is None


def test_nan():
    """
    Floating point nan is decoded correctly.
    """
    serialized = cbor2.dumps(math.nan)
    result = ANY_SCHEMA.validate_cbor(serialized, True)
    assert math.isnan(result)


def test_sets():
    """
    Sets can be decoded.
    """
    structure = {1, b"abc", "def", 2.3}
    serialized = cbor2.dumps(structure)
    result = ANY_SCHEMA.validate_cbor(serialized, True)
    assert result == structure


@given(data=st.binary(max_size=10_000_000))
def test_large_bytes(data):
    """
    Deserialization of large bytes.
    """
    serialized = cbor2.dumps(data)
    assert ANY_SCHEMA.validate_cbor(serialized, True) == data



"""Tests for pycddl."""

import platform
import resource
import mmap

from psutil import Process
import pytest
import cbor2

from pycddl import Schema, ValidationError

from .utils import BSTR_SCHEMA, BSTR_100M, BSTR_1K


def assert_invalid_caught(schema, data):
    """
    The schema correctly identifies that data as invalid.
    """
    with pytest.raises(ValidationError):
        schema.validate_cbor(cbor2.dumps(data))


def test_invalid_schema_errors_out():
    """
    Attempting to create a new ``CDDLSchema`` with an invalid CDDL schema
    results in a ValueError.
    """
    with pytest.raises(ValueError):
        Schema(
            """
    reputation-object = {
        application: text
        reputons: [* reputon]

    """
        )


REPUTON_SCHEMA = """\
reputation-object = {
  application: text
  reputons: [* reputon]
}

reputon = {
  rater: text
  assertion: text
  rated: text
  rating: float16
  ? confidence: float16
  ? normal-rating: float16
  ? sample-size: uint
  ? generated: uint
  ? expires: uint
  * text => any
}
"""


def test_schema_validates_good_document():
    """
    A valid schema will validate a valid document (i.e. no exception is
    raised).
    """
    schema = Schema(REPUTON_SCHEMA)
    for document in [
        {"application": "blah", "reputons": []},
        {
            "application": "conchometry",
            "reputons": [
                {
                    "rater": "Ephthianura",
                    "assertion": "codding",
                    "rated": "sphaerolitic",
                    "rating": 0.34133473256800795,
                    "confidence": 0.9481983064298332,
                    "expires": 1568,
                    "unplaster": "grassy",
                },
                {
                    "rater": "nonchargeable",
                    "assertion": "raglan",
                    "rated": "alienage",
                    "rating": 0.5724646875815566,
                    "sample-size": 3514,
                    "Aldebaran": "unchurched",
                    "puruloid": "impersonable",
                    "uninfracted": "pericarpoidal",
                    "schorl": "Caro",
                },
            ],
        },
    ]:
        schema.validate_cbor(cbor2.dumps(document))


def test_schema_fails_bad_documents():
    """
    Bad documents cause ``validate_cbor()`` to raise a ``ValidationError``.
    """
    schema = Schema(REPUTON_SCHEMA)
    for bad_document in [
        b"",
        cbor2.dumps({"application": "blah"}),  # missing reputons key
        cbor2.dumps({"application": "blah", "reputons": "NOT A LIST"}),
    ]:
        with pytest.raises(ValidationError):
            schema.validate_cbor(bad_document)


def test_integer_value_enforcement():
    """
    Schemas that limit minimum integer value are enforced.  This is important
    for security, for example.
    """
    uint_schema = Schema(
        """
    object = {
        xint: uint
    }
    """
    )
    for i in [0, 1, 4, 5, 500, 1000000]:
        uint_schema.validate_cbor(cbor2.dumps({"xint": i}))
    for i in [-1, -10000, "x", 0.3]:
        assert_invalid_caught(uint_schema, i)

    more_than_3_schema = Schema(
        """
    object = {
        xint: int .gt 3
    }
    """
    )
    for i in [4, 5, 500, 1000000]:
        more_than_3_schema.validate_cbor(cbor2.dumps({"xint": i}))
    for i in [-1, -10000, "x", 0.3, 0, 1, 2, 3]:
        assert_invalid_caught(more_than_3_schema, {"xint": i})


def test_schema_repr():
    """``repr(Schema)`` reflects the schema string."""
    schema_text = """
    object = {
        xint: int .gt 3
        "value": uint
    }
    """
    schema = Schema(schema_text)
    assert repr(schema) == f'Schema("""{schema_text}""")'


def get_max_rss_mb():
    # ru_maxrss is kilobytes on Linux, bytes on macOS
    result = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024
    if platform.system() == "Darwin":
        result /= 1024
    return result


@pytest.mark.skipif(platform.system() == "Windows", reason="Need POSIX to check maxrss")
def test_memory_usage():
    """
    Validating a large document doesn't significantly increase memory usage.
    """
    process = Process()
    # This returns units of bytes, so divide appropriately:
    before_rss_mb = process.memory_info().rss / (1024 * 1024)
    before_max_rss_mb = get_max_rss_mb()

    # If this fails, this test won't be meaningful because new allocation won't
    # budge max RSS. So if this fails, to fix it you should try to adjust
    # runtime to ensure we don't have extra high max RSS. E.g. in utils.py we
    # try to generate large CBOR docs in a subprocess.
    #
    # TODO Fil should really expose an API for "current in-use memory and
    # current peak memory", it has the info...
    assert (
        before_max_rss_mb - before_rss_mb < 105
    ), "This is an environmental check; see code"

    BSTR_SCHEMA.validate_cbor(BSTR_100M)
    new_max_rss_mb = get_max_rss_mb()

    # We're validating a 100MB document. The underlying parser used by the Rust
    # library allocates memory linearly based on input (see
    # https://github.com/anweiss/cddl/issues/167), but we at least should not
    # be copying memory much beyond that. Ideally this should be more like < 10.
    assert new_max_rss_mb - before_rss_mb < 120


def test_buffer_interface(tmp_path):
    """
    It's possible to pass in read-only buffers, e.g. ``memoryview``.
    """
    # memoryview() of bytes works fine: it's read-only.
    BSTR_SCHEMA.validate_cbor(memoryview(BSTR_1K))

    # But if we mmap() an existing file in read-only way, that's fine:
    path = tmp_path / "out.cbor"
    with path.open("wb") as f:
        f.write(BSTR_1K)
    with path.open("rb") as f2:
        mapping = mmap.mmap(f2.fileno(), len(BSTR_1K), access=mmap.ACCESS_READ)
        BSTR_SCHEMA.validate_cbor(mapping)


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="PyPy is buggy, it thinks bytearray() is read-only...",
)
def test_buffer_interface_refuses_writeable():
    """
    Writeable buffers won't get validated.
    """
    # bytearray won't work because it's mutable:
    arr = bytearray(BSTR_1K)
    with pytest.raises(ValueError):
        BSTR_SCHEMA.validate_cbor(arr)

    # Likewise for writeable mmap():
    mapping = mmap.mmap(-1, len(BSTR_1K))
    mapping[:] = BSTR_1K
    with pytest.raises(ValueError):
        BSTR_SCHEMA.validate_cbor(mapping)



"""
Benchmarks for parsing.

Run with pytest:

    $ pytest benchmarks.py
"""

import pytest

from .utils import BSTR_1K, BSTR_1M, BSTR_100M, BSTR_SCHEMA


@pytest.mark.parametrize("document_length", [1000, 1_000_000])
def test_validate_bytestring(benchmark, document_length, tmp_path):
    """
    Measure how long it takes to validate a bytestring of a given size.

    The hope is that validation time is still fast for large byte strings.
    """
    document = {1000: BSTR_1K, 1_000_000: BSTR_1M, 100_000_000: BSTR_100M}[
        document_length
    ]
    benchmark(lambda: BSTR_SCHEMA.validate_cbor(document))



__all__ = ["Schema", "ValidationError"]

import platform
from .pycddl import Schema, ValidationError

if platform.python_implementation() == "PyPy":
    _Schema = Schema

    class Schema:
        def __init__(self, schema_string):
            self._schema = _Schema(schema_string)

        def validate_cbor(self, cbor, deserialize=False):
            # On PyPy, not wrapping with memoryview() results in bytes being
            # copied, which is very memory inefficient for large documents.
            return self._schema.validate_cbor(memoryview(cbor), deserialize)

        def __repr__(self):
            return repr(self._schema)

        def __str__(self):
            return str(self._schema)

    Schema.__doc__ = _Schema.__doc__
    Schema.validate_cbor.__doc__ = Schema.validate_cbor.__doc__
